# Adapted from Xu et al. @ ICML 2021 and 2023
from time import time
import pandas as pd
import numpy as np

from .utils_SPCI import (
    binning,
    generate_bootstrap_samples,
    strided_app,
)

# Only used:
# SPCI_and_EnbPI
# fit_bootstrap_models_online_multistep
# compute_PIs_Ensemble_online
# Attributes
# Ensemble_pred_interval_centers: point forecast
# PIs_Ensemble: predictions intervals PIs_Ensemble['lower'], PIs_Ensemble['upper']

#### Main Class ####
class SPCI_and_EnbPI:
    """
        Create prediction intervals assuming Y_t = f(X_t) + \sigma(X_t)\eps_t
        Currently, assume the regression function is by default MLP implemented with PyTorch, as it needs to estimate BOTH f(X_t) and \sigma(X_t), where the latter is impossible to estimate using scikit-learn modules

        Most things carry out, except that we need to have different estimators for f and \sigma.

        fit_func = None: use MLP above
    """

    def __init__(self, X_train, Y_train, h=5, fit_func=None):

        self.regressor = fit_func
        self.X_train = X_train
        self.Y_train = Y_train

        # Predicted training data centers by EnbPI
        n, n1 = self.X_train.shape[0], h
        self.Ensemble_train_interval_centers = np.ones(n) * np.inf
        self.Ensemble_train_interval_sigma = np.ones(n) * np.inf
        # Predicted test data centers by EnbPI
        self.Ensemble_pred_interval_centers = np.ones(n1) * np.inf
        self.Ensemble_pred_interval_sigma = np.ones(n1) * np.inf
        self.Ensemble_online_resid = np.ones(n + n1) * np.inf  # LOO scores
        self.beta_hat_bins = []

        #### Other hyperparameters for training (mostly simulation) ####
        # QRF training & how it treats the samples
        self.weigh_residuals = False  # Whether we weigh current residuals more.
        self.c = (
            0.995  # If self.weight_residuals, weights[s] = self.c ** s, s\geq 0
        )
        self.n_estimators = 10  # Num trees for QRF
        self.max_d = 2  # Max depth for fitting QRF
        self.criterion = "mse"  # 'mse' or 'mae'
        # search of \beta^* \in [0,\alpha]
        self.bins = 5  # break [0,\alpha] into bins
        # how many LOO training residuals to use for training current QRF
        self.T1 = None  # None = use all

        # additional attributes
        self.d = None
        self.fit_sigmaX = None
        self.alpha = None
        self.past_window = None
        self.train_idx = None
        self.b = None
        self.PIs_Ensemble = None
        self.test_idx = None
        self.h = None

    def one_boot_prediction(self, X_boot, Y_boot, X_full):
        # NOTE, NO sigma estimation because these methods by deFAULT are fitting Y, but we have no observation of errors
        model = self.regressor
        model.fit(X_boot, Y_boot)
        boot_fX_pred = model.predict(X_full, h=self.h)
        boot_sigma_pred = 0
        return boot_fX_pred, boot_sigma_pred

    def fit_bootstrap_models_online_multistep(
        self, B=100, fit_sigmaX=True, stride=1, h=5
    ):
        """
          Train B bootstrap estimators from subsets of (X_train, Y_train), compute aggregated predictors, and compute the residuals
          fit_sigmaX: If False, just avoid predicting \sigma(X_t) by defaulting it to 1

          stride: int. If > 1, then we perform multi-step prediction, where we have to fit stride*B boostrap predictors.
            Idea: train on (X_i,Y_i), i=1,...,n-stride
            Then predict on X_1,X_{1+s},...,X_{1+k*s} where 1+k*s <= n+n1
            Note, when doing LOO prediction thus only care above the points above

          h: int. forecasting horizon
        """
        n, self.d = self.X_train.shape
        self.fit_sigmaX = fit_sigmaX
        self.h = h
        n1 = h  # forecasting horizon
        N = n - stride + 1  # Total training data each one-step predictor sees
        # We make prediction every s step ahead, so these are feature the model sees
        train_pred_idx = np.arange(0, n, stride)
        # We make prediction every s step ahead, so these are feature the model sees
        test_pred_idx = np.arange(n, n + n1, stride)
        self.train_idx = train_pred_idx
        self.test_idx = test_pred_idx
        # Only contains features that are observed every stride steps
        X_full = self.X_train[train_pred_idx, :]
        nsub, n1sub = len(train_pred_idx), len(test_pred_idx)
        for s in range(stride):
            # 1 - Create containers for predictions
            # hold indices of training data for each f^b
            boot_samples_idx = generate_bootstrap_samples(N, N, B)
            # for i^th column, it shows which f^b uses i in training (so exclude in aggregation)
            in_boot_sample = np.zeros((B, N), dtype=bool)
            # hold predictions from each f^b for fX and sigma&b for sigma
            boot_predictionsFX = np.zeros((B, nsub + n1sub))
            boot_predictionsSigmaX = np.ones((B, nsub + n1sub))
            # We actually would just use n1sub rows, as we only observe this number of features
            out_sample_predictFX = np.zeros((n, n1sub))
            out_sample_predictSigmaX = np.ones((n, n1sub))

            # 2 - Start bootstrap prediction
            start = time()
            for b in range(B):
                self.b = b
                X_boot, Y_boot = (
                    self.X_train[boot_samples_idx[b], :],
                    self.Y_train[s : s + N][boot_samples_idx[b],],
                )
                in_boot_sample[b, boot_samples_idx[b]] = True
                boot_fX_pred, boot_sigma_pred = self.one_boot_prediction(
                    X_boot, Y_boot, X_full
                )
                boot_predictionsFX[b] = boot_fX_pred
                if self.fit_sigmaX:
                    boot_predictionsSigmaX[b] = boot_sigma_pred
            print(
                f"{s+1}/{stride} multi-step: finish Fitting {B} Bootstrap models, took {time()-start} secs."
            )

            # 3 - Obtain LOO residuals (train and test) and prediction for test data '''
            start = time()
            # Consider LOO, but here ONLY for the indices being predicted
            for j, i in enumerate(train_pred_idx):
                # j: counter and i: actual index X_{0+j*stride}
                if i < N:
                    b_keep = np.argwhere(~(in_boot_sample[:, i])).reshape(-1)
                    if len(b_keep) == 0:
                        # All bootstrap estimators are trained on this model
                        b_keep = 0  # More rigorously, it should be None, but in practice, the difference is minor
                else:
                    # This feature is not used in training, but used in prediction
                    b_keep = range(B)
                pred_iFX = boot_predictionsFX[b_keep, j].mean()
                pred_iSigmaX = boot_predictionsSigmaX[b_keep, j].mean()
                pred_testFX = boot_predictionsFX[b_keep, nsub:].mean(0)
                pred_testSigmaX = boot_predictionsSigmaX[b_keep, nsub:].mean(0)
                # Populate the training prediction
                # We add s because of multi-step procedure, so f(X_t) is for Y_t+s
                true_idx = min(i + s, n - 1)
                self.Ensemble_train_interval_centers[true_idx] = pred_iFX
                self.Ensemble_train_interval_sigma[true_idx] = pred_iSigmaX
                resid_LOO = (self.Y_train[true_idx] - pred_iFX) / pred_iSigmaX
                out_sample_predictFX[i] = pred_testFX
                out_sample_predictSigmaX[i] = pred_testSigmaX
                self.Ensemble_online_resid[true_idx] = resid_LOO.item()
            sorted_out_sample_predictFX = out_sample_predictFX[
                train_pred_idx
            ].mean(
                0
            )  # length ceil(n1/stride)
            sorted_out_sample_predictSigmaX = out_sample_predictSigmaX[
                train_pred_idx
            ].mean(
                0
            )  # length ceil(n1/stride)
            pred_idx = np.minimum(test_pred_idx - n + s, n1 - 1)
            self.Ensemble_pred_interval_centers[
                pred_idx
            ] = sorted_out_sample_predictFX
            self.Ensemble_pred_interval_sigma[
                pred_idx
            ] = sorted_out_sample_predictSigmaX
            # pred_full_idx = np.minimum(test_pred_idx + s, n + n1 - 1)
            # resid_out_sample = (self.Y_predict[pred_idx] - sorted_out_sample_predictFX) / sorted_out_sample_predictSigmaX
            # self.Ensemble_online_resid[pred_full_idx] = resid_out_sample
        # Sanity check
        # num_inf = (self.Ensemble_online_resid == np.inf).sum()
        # if num_inf > 0:
        #     # print(
        #     #    f'Something can be wrong, as {num_inf}/{n+n1} residuals are not all computed')
        #     print(
        #         f"Something can be wrong, as {num_inf}/{n} residuals are not all computed"
        #     )
        #     print(np.where(self.Ensemble_online_resid == np.inf))

    def compute_PIs_Ensemble_online(
        self,
        alpha,
        stride=1,
        smallT=True,
        past_window=100,
        use_SPCI=False,
        quantile_regr="RF",
    ):
        """
            stride: control how many steps we predict ahead
            smallT: if True, we would only start with the last n number of LOO residuals, rather than use the full length T ones. Used in change detection
                NOTE: smallT can be important if time-series is very dynamic, in which case training MORE data may actaully be worse (because quantile longer)
                HOWEVER, if fit quantile regression, set it to be FALSE because we want to have many training pts for the quantile regressor
            use_SPCI: if True, we fit conditional quantile to compute the widths, rather than simply using empirical quantile
        """
        self.alpha = alpha
        n1 = self.X_train.shape[0]
        self.past_window = (
            past_window  # For SPCI, this is the "lag" for predicting quantile
        )
        if smallT:
            # Namely, for special use of EnbPI, only use at most past_window number of LOO residuals.
            n1 = min(self.past_window, self.X_train.shape[0])
        # Now f^b and LOO residuals have been constructed from earlier
        out_sample_predict = self.Ensemble_pred_interval_centers
        out_sample_predictSigmaX = self.Ensemble_pred_interval_sigma
        start = time()
        # Matrix, where each row is a UNIQUE slice of residuals with length stride.
        if use_SPCI:
            s = stride
            stride = 1
        # NOTE, NOT ALL rows are actually "observable" in multi-step context, as this is rolling
        resid_strided = strided_app(
            self.Ensemble_online_resid[len(self.X_train) - n1 : -1], n1, stride
        )
        print(f"Shape of slided residual lists is {resid_strided.shape}")
        num_unique_resid = resid_strided.shape[0]
        width_left = np.zeros(num_unique_resid)
        width_right = np.zeros(num_unique_resid)
        # # NEW, alpha becomes alpha_t. Uncomment things below if we decide to use this upgraded EnbPI
        # alpha_t = alpha
        # errs = []
        # gamma = 0.005
        # method = 'simple'  # 'simple' or 'complex'
        # self.alphas = []
        # NOTE: 'max_features='log2', max_depth=2' make the model "simpler", which improves performance in practice
        self.QRF_ls = []
        self.i_star_ls = []
        for i in range(num_unique_resid):
            curr_SigmaX = out_sample_predictSigmaX[i].item()
            if use_SPCI:
                remainder = i % s
                if remainder == 0:
                    # Update QRF
                    past_resid = resid_strided[i, :]
                    n2 = self.past_window
                    resid_pred = self.multi_step_QRF(past_resid, i, s, n2)
                # Use the fitted regressor.
                # NOTE, residX is NOT the same as before, as it depends on
                # "past_resid", which has most entries replaced.
                rfqr = self.QRF_ls[remainder]
                i_star = self.i_star_ls[remainder]
                wid_all = rfqr.predict(resid_pred)
                num_mid = int(len(wid_all) / 2)
                wid_left = wid_all[i_star]
                wid_right = wid_all[num_mid + i_star]
                width_left[i] = curr_SigmaX * wid_left
                width_right[i] = curr_SigmaX * wid_right
                num_print = int(num_unique_resid / 20)
                if num_print == 0:
                    print(
                        f"Width at test {i} is {width_right[i]-width_left[i]}"
                    )
                else:
                    if i % num_print == 0:
                        print(
                            f"Width at test {i} is {width_right[i]-width_left[i]}"
                        )
            else:  # use EnbPI
                past_resid = resid_strided[i, :]
                # Naive empirical quantile, where we use the SAME residuals for multi-step prediction
                # The number of bins will be determined INSIDE binning
                beta_hat_bin = binning(past_resid, alpha)
                # beta_hat_bin = binning(past_resid, alpha_t)
                self.beta_hat_bins.append(beta_hat_bin)
                # width_left[i] = curr_SigmaX * np.percentile(
                #     past_resid, math.ceil(100 * beta_hat_bin)
                # )
                # width_right[i] = curr_SigmaX * np.percentile(
                #     past_resid, math.ceil(100 * (1 - alpha + beta_hat_bin))
                # )
        print(
            f"Finish Computing {num_unique_resid} UNIQUE Prediction Intervals, took {time()-start} secs."
        )
        # Ntest = len(out_sample_predict)
        # This is because |width|=T1/stride.
        # width_left = np.repeat(width_left, stride)[:Ntest]
        # This is because |width|=T1/stride.
        # width_right = np.repeat(width_right, stride)[:Ntest]
        PIs_Ensemble = pd.DataFrame(
            np.c_[
                out_sample_predict + width_left,
                out_sample_predict + width_right,
            ],
            columns=["lower", "upper"],
        )
        self.PIs_Ensemble = PIs_Ensemble



from .lazyClassifier import LazyClassifier
from .lazyRegressor import LazyRegressor
from .lazyMTS import LazyMTS
from .lazydeepClassifier import LazyDeepClassifier
from .lazydeepRegressor import LazyDeepRegressor

__all__ = ["LazyClassifier", "LazyRegressor", "LazyMTS", 
"LazyDeepClassifier", "LazyDeepRegressor"]


# Adapted from Xu et al. @ ICML 2021 and 2023

# from statsmodels.regression.quantile_regression import QuantReg
# from sklearn.linear_model import QuantileRegressor
import numpy as np
import math
import pandas as pd


#### From utils_EnbPI ####
def adjust_alpha_t(alpha_t, alpha, errs, gamma=0.005, method="simple"):
    if method == "simple":
        # Eq. (2) of Adaptive CI
        return alpha_t + gamma * (alpha - errs[-1])
    else:
        # Eq. (3) of Adaptive CI with particular w_s as given
        t = len(errs)
        errs = np.array(errs)
        w_s_ls = np.array(
            [0.95 ** (t - i) for i in range(t)]
        )  # Furtherest to Most recent
        return alpha_t + gamma * (alpha - w_s_ls.dot(errs))


def ave_cov_width(df, Y):
    coverage_res = (
        (np.array(df["lower"]) <= Y) & (np.array(df["upper"]) >= Y)
    ).mean()
    print(f"Average Coverage is {coverage_res}")
    width_res = (df["upper"] - df["lower"]).mean()
    print(f"Average Width is {width_res}")
    return [coverage_res, width_res]


#### Miscellaneous ####


window_size = 300


def rolling_avg(x, window=window_size):
    return np.convolve(x, np.ones(window) / window)[(window - 1) : -window]


def dict_to_latex(dict, train_ls):
    DF = pd.DataFrame.from_dict(np.vstack(dict.values()))
    keys = list(dict.keys())
    index = np.array(
        [[f"{key} coverage", f"{key} width"] for key in keys]
    ).flatten()
    DF.index = index
    DF.columns = train_ls
    print(DF)
    print(DF.round(2).to_latex())


def make_NP_df(X, Y):
    Xnames = [f"X{a}" for a in np.arange(X.shape[1]).astype(str)]
    full_names = ["ds"] + Xnames + ["y"]
    date_tmp = pd.date_range(start="1/1/2018", periods=len(Y)).astype(
        str
    )  # Artificial
    df_tmp = pd.DataFrame(np.c_[date_tmp, X, Y], columns=full_names)
    return df_tmp, Xnames


def generate_bootstrap_samples(n, m, B):
    """
      Return: B-by-m matrix, where row b gives the indices for b-th bootstrap sample
    """
    samples_idx = np.zeros((B, m), dtype=int)
    for b in range(B):
        sample_idx = np.random.choice(n, m)
        samples_idx[b, :] = sample_idx
    return samples_idx


def strided_app(a, L, S):  # Window len = L, Stride len/stepsize = S
    nrows = ((a.size - L) // S) + 1
    n = a.strides[0]
    return np.lib.stride_tricks.as_strided(
        a, shape=(nrows, L), strides=(S * n, n)
    )


def binning(past_resid, alpha):
    """
    Input:
        past residuals: evident
        alpha: signifance level
    Output:
        beta_hat_bin as argmin of the difference
    Description:
        Compute the beta^hat_bin from past_resid, by breaking [0,alpha] into bins (like 20). It is enough for small alpha
        number of bins are determined rather automatic, relative the size of whole domain
    """
    bins = 5  # For computation, can just reduce it to like 10 or 5 in real data
    beta_ls = np.linspace(start=0, stop=alpha, num=bins)
    width = np.zeros(bins)
    for i in range(bins):
        width[i] = np.percentile(
            past_resid, math.ceil(100 * (1 - alpha + beta_ls[i]))
        ) - np.percentile(past_resid, math.ceil(100 * beta_ls[i]))
    i_star = np.argmin(width)
    return beta_ls[i_star]


def binning_use_RF_quantile_regr(
    quantile_regr, Xtrain, Ytrain, feature, beta_ls, sample_weight=None
):
    # API ref: https://sklearn-quantile.readthedocs.io/en/latest/generated/sklearn_quantile.RandomForestQuantileRegressor.html
    feature = feature.reshape(1, -1)
    low_high_pred = quantile_regr.fit(Xtrain, Ytrain, sample_weight).predict(
        feature
    )
    num_mid = int(len(low_high_pred) / 2)
    low_pred, high_pred = low_high_pred[:num_mid], low_high_pred[num_mid:]
    width = (high_pred - low_pred).flatten()
    i_star = np.argmin(width)
    wid_left, wid_right = low_pred[i_star], high_pred[i_star]
    return i_star, beta_ls[i_star], wid_left, wid_right


def merge_table_mean_std(table_result, colnames=None):
    M, N = table_result.shape[0], int(table_result.shape[1] / 2)
    table = np.zeros((M, N), dtype=object)
    idx = table_result.index
    for i in range(M):
        for j in range(N):
            table[
                i, j
            ] = f"{table_result.iloc[i,2*j]} ({table_result.iloc[i,2*j+1]})"
    colnames = np.array(
        [[f"{name} coverage", f"{name} width"] for name in colnames]
    ).flatten()
    return pd.DataFrame(table, index=idx, columns=colnames)




from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.utils import all_estimators

removed_classifiers = [
    "ClassifierChain",
    "ComplementNB",
    "MultiOutputClassifier", 
    "MultinomialNB", 
    "OneVsOneClassifier",
    "OneVsRestClassifier",
    "OutputCodeClassifier",
    "RadiusNeighborsClassifier",
    "VotingClassifier",
]

removed_regressors = [
    "TheilSenRegressor",
    "ARDRegression", 
    "CCA", 
    "IsotonicRegression", 
    "StackingRegressor",
    "MultiOutputRegressor", 
    "MultiTaskElasticNet", 
    "MultiTaskElasticNetCV", 
    "MultiTaskLasso", 
    "MultiTaskLassoCV", 
    "PLSCanonical", 
    "PLSRegression", 
    "RadiusNeighborsRegressor", 
    "RegressorChain", 
    "VotingRegressor", 
]

CLASSIFIERS = [
    est
    for est in all_estimators()
    if (issubclass(est[1], ClassifierMixin) and (est[0] not in removed_classifiers))
]

REGRESSORS = [
    est
    for est in all_estimators()
    if (issubclass(est[1], RegressorMixin) and (est[0] not in removed_regressors))
]



# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import numpy as np
from tqdm import tqdm
from ..sampling.helpers import dosubsample


# 0 - utils -----

# 0 - 0 data structures & funcs -----

def call_f(f, x):
    return f(x)

def calc_grad(f, x):
    return numerical_gradient(f, x)
    
def calc_hessian(f, x):
    return numerical_hessian(f, x)       

def generate_index(response, batch_prop=1.0, 
                   randomization="strat", seed=123):
    """Generation of indices for Stochastic gradient descent."""
    
    n = len(response)    

    if batch_prop < 1:
    
        if randomization=="strat":        
            return dosubsample(response, batch_prop, seed=seed)
        
        if randomization=="shuffle":                    
            np.random.seed(seed)        
            return np.asarray(np.floor(np.random.choice(range(n), 
                                    size=np.floor(batch_prop*n), 
                                    replace=False)), dtype=np.int)   
    return None


def check_is_min(f, x):
    return (numerical_hessian(f, x)[1])*1


def calc_fplus(x, alpha, p_k):        
    
    p = len(x)
    res = np.zeros(p)     
    j = 0
    
    for j in range(p):
        res[j] = x[j] + alpha*p_k[j]
        
    return res 

# Useful functions -----

# 1 - gradient-----
        
def numerical_gradient(f, x, **kwargs):
    
    p = len(x)
    ix = 0    
    eps_factor = 6.055454452393343e-06 # machine eps ** (1/3)
    zero = 2.220446049250313e-16    
    res = np.zeros_like(x)
    
    f_ = lambda x: f(x, **kwargs)
       
    for ix in range(p):
        
        value_x = x[ix]       
        
        h = max(eps_factor*value_x, 1e-8)
        
        x[ix] = value_x + h
        fx_plus = call_f(f_, x)
        
        x[ix] = value_x - h
        fx_minus = call_f(f_, x)
        
        x[ix] = value_x # restore
        
        res[ix] = (fx_plus - fx_minus)/(2*h)        
                    
    return np.asarray(res)


# 2 - hessian-----
    
def numerical_hessian(f, x, **kwargs):
    
    p = len(x)
    ix = 0
    jx = 0    
    eps_factor = 0.0001220703125 # machine eps ** (1/4)
    zero = 2.220446049250313e-16
    H = np.zeros((p, p))
    temp = 0 
    
    f_ = lambda x: f(x, **kwargs)
    
    fx = call_f(f_, x)
       
    for ix in range(p):
        
        for jx in range(ix, p):
            
            if (ix < jx):
                
                value_x = x[ix]            
                value_y = x[jx]
                
                h = max(eps_factor*value_x, 1e-8)
                k = max(eps_factor*value_y, 1e-8) 
                
                x[ix] = value_x + h
                x[jx] = value_y + k
                fx_plus = call_f(f_, x)
                
                x[ix] = value_x + h
                x[jx] = value_y - k
                fx_plus_minus = call_f(f_, x)
                
                x[ix] = value_x - h
                x[jx] = value_y + k
                fx_minus_plus = call_f(f_, x)
                
                x[ix] = value_x - h
                x[jx] = value_y - k
                fx_minus = call_f(f_, x)
                
                x[ix] = value_x # restore
                x[jx] = value_y # restore
                
                temp = (fx_plus - fx_plus_minus - fx_minus_plus + fx_minus)/(4*h*k)        
                                            
            else:
                
                value_x = x[ix]                                                            
                
                h = max(eps_factor*value_x, 1e-8)
                
                x[ix] = value_x + h
                fx_plus = call_f(f_, x)
                                
                x[ix] = value_x - h
                fx_minus = call_f(f_, x)
                
                x[ix] = value_x # restore
                
                temp = (fx_plus - 2*fx + fx_minus)/(h**2)        
                
            H[ix, jx] = temp            
            H[jx, ix] = temp                
                
    res = np.asarray(H)
                    
    return res, all(np.linalg.eig(res)[0] > 0)


# 3 - One-hot encoder -----

def one_hot_encode(y, n_classes):
    
    n_obs = len(y)
    res = np.zeros((n_obs, n_classes), dtype=float)        

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)


# Coordinate descent (Stochastic) -----

# 1 - algos -----

def scd(loss_func, response, x, num_iters=200, 
        batch_prop=1.0, learning_rate=0.01, mass=0.9, 
        decay=0.1, method="momentum", randomization="strat", 
        tolerance=1e-3, verbose=1, 
        **kwargs):
    """Stochastic coordinate descent with momentum and adaptive learning rates."""
    
    i = 0 
    j = 0
    n = len(response)
    p = len(x)
    velocity = np.zeros(p)
    losses = []

    if verbose == 1:
        iterator = tqdm(range(num_iters))          
    else:
        iterator = range(num_iters)
    
    f = lambda x: loss_func(x, **kwargs)              
       
    if method == "momentum":                

        for i in iterator: 
                   
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)   

            def f_j(h, xx, j):        
                value_x = 0        
                res = 0 
                value_x = xx[j]            
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)          
                xx[j] = value_x
                return res                                                                                     
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            for j in range(p):                          
                h0 = 6.055454452393343e-06*x[j]        
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j))/(2*h0)        
                velocity[j] = mass * velocity[j] - learning_rate * grad_x  
                x[j] = x[j] + velocity[j]        
            
            diff += np.asarray(x)  

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")
            
            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
            
            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try: 
                    print(np.linalg.norm(diff, 1)) 
                except:
                    pass   

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])     
    
    if method in ("exp", "poly"):
        
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def f_j(h, xx, j):        
                value_x = 0        
                res = 0 
                value_x = xx[j]            
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)          
                xx[j] = value_x
                return res      
                                                
            decay_rate = (1 + decay*i) if (method == "poly") else np.exp(decay*i)
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            losses.append(f(x))
            
            for j in range(p):  
                h0 = 6.055454452393343e-06*x[j]        
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j))/(2*h0)        
                x[j] = x[j] - grad_x*learning_rate/decay_rate
                
            diff += np.asarray(x) 

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}") 
            
            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break

            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))   
                except:
                    pass 

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

                    
    return np.asarray(x), num_iters, losses


# Gradient descent (Stochastic) -----  

def sgd(loss_func, response, x, num_iters=200, 
        batch_prop=1.0, learning_rate=0.01, mass=0.9, 
        decay=0.1, method="momentum", randomization="strat", 
        tolerance=1e-3, verbose=1, 
        **kwargs):
    """Stochastic gradient descent with momentum and adaptive learning rates."""
    
    i = 0 
    j = 0
    n = len(response)
    p = len(x)
    velocity = np.zeros(p)
    grad_i = np.zeros(p)
    losses = []
    
    if verbose == 1:
        iterator = tqdm(range(num_iters))          
    else:
        iterator = range(num_iters)
            
    f = lambda x: loss_func(x, **kwargs)                            
    
    if method == "momentum":                
    
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def objective(x):
                return loss_func(x, row_index=idx, **kwargs)
            
            #grad_i = numerical_gradient(objective, x)                        
            grad_i = calc_grad(objective, x)    
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            for j in range(p):                
                velocity[j] = mass * velocity[j] - learning_rate * grad_i[j]                        
                x[j] = x[j] + velocity[j]
            
            diff += np.asarray(x)
            
            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
            
            if verbose == 2:       

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))                            
                except:
                    pass  

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    
    if method in ("exp", "poly"):
        
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def objective(x):
                return loss_func(x, row_index=idx, **kwargs)
            
            #grad_i = numerical_gradient(objective, x)
            grad_i = calc_grad(objective, x)            
                        
            decay_rate = (1 + decay*i) if (method == "poly") else np.exp(decay*i)
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            for j in range(p):   

                x[j] = x[j] - grad_i[j]*learning_rate/decay_rate

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            diff += np.asarray(x)  
            
            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
        
            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")
                
                try:
                    print(np.linalg.norm(diff, 1))   
                except:
                    pass
                
                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])
    
    return np.asarray(x), num_iters, losses



# Adapted from: https://github.com/shankarpandala/lazypredict
# Author: Shankar Rao Pandala <shankar.pandala@live.com>

import numpy as np
import pandas as pd
from tqdm import tqdm
import time
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import ClassifierMixin
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    roc_auc_score,
    f1_score,
)
from .config import CLASSIFIERS
from ..custom import Custom, CustomClassifier

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

categorical_transformer_low = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        ("encoding", OneHotEncoder(handle_unknown="ignore", sparse=False)),
    ]
)

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


def get_card_split(df, cols, n=11):
    """
    Splits categorical columns into 2 lists based on cardinality (i.e # of unique values)
    Parameters
    ----------
    df : Pandas DataFrame
        DataFrame from which the cardinality of the columns is calculated.
    cols : list-like
        Categorical columns to list
    n : int, optional (default=11)
        The value of 'n' will be used to split columns.
    Returns
    -------
    card_low : list-like
        Columns with cardinality < n
    card_high : list-like
        Columns with cardinality >= n
    """
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


class LazyClassifier(Custom, ClassifierMixin):
    """
    This module helps in fitting to all the classification algorithms that are available in Scikit-learn to nnetsauce's CustomClassifier
    Parameters
    ----------
    verbose : int, optional (default=0)
        For the liblinear and lbfgs solvers set verbose to any positive
        number for verbosity.
    ignore_warnings : bool, optional (default=True)
        When set to True, the warning related to algorigms that are not able to run are ignored.
    custom_metric : function, optional (default=None)
        When function is provided, models are evaluated based on the custom evaluation metric provided.
    prediction : bool, optional (default=False)
        When set to True, the predictions of all the models models are returned as dataframe.
    classifiers : list, optional (default="all")
        When function is provided, trains the chosen classifier(s).

    Examples
    --------
    >>> import nnetsauce as ns 
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.model_selection import train_test_split
    >>> data = load_breast_cancer()
    >>> X = data.data
    >>> y= data.target
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.2, random_state=123)
    >>> clf = ns.LazyClassifier(verbose=0, ignore_warnings=True, custom_metric=None)
    >>> models, predictions = clf.fit(X_train, X_test, y_train, y_test)
    >>> model_dictionary = clf.provide_models(X_train,X_test,y_train,y_test)
    >>> print(models)
    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        random_state=42,
        classifiers="all",
        preprocess=False,
        # CustomClassifier attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.models = {}
        self.random_state = random_state
        self.classifiers = classifiers
        self.preprocess = preprocess
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

    def fit(self, X_train, X_test, y_train, y_test):
        """Fit Classification algorithms to X_train and y_train, predict and score on X_test, y_test.
        Parameters
        ----------
        X_train : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        X_test : array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.
        y_train : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y_test : array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.
        Returns
        -------
        scores : Pandas DataFrame
            Returns metrics of all the models in a Pandas DataFrame.
        predictions : Pandas DataFrame
            Returns predictions of all the models in a Pandas DataFrame.
        """
        Accuracy = []
        B_Accuracy = []
        ROC_AUC = []
        F1 = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric is not None:
            CUSTOM_METRIC = []

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess is True:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )

        if self.classifiers == "all":
            self.classifiers = CLASSIFIERS
        else:
            try:
                temp_list = []
                for classifier in self.classifiers:
                    full_name = (classifier.__name__, classifier)
                    temp_list.append(full_name)
                self.classifiers = temp_list
            except Exception as exception:
                print(exception)
                print("Invalid Classifier(s)")

        if self.preprocess is True:

            for name, model in tqdm(self.classifiers):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = Pipeline(
                            [
                                ("preprocessor", preprocessor),
                                (
                                    "classifier",
                                    CustomClassifier(
                                        obj=model(
                                            random_state=self.random_state
                                        ),
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        col_sample=self.col_sample,
                                        row_sample=self.row_sample,
                                        seed=self.seed,
                                        backend=self.backend,
                                    ),
                                ),
                            ]
                        )
                    else:
                        pipe = Pipeline(
                            [
                                ("preprocessor", preprocessor),
                                (
                                    "classifier",
                                    CustomClassifier(
                                        obj=model(),
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        col_sample=self.col_sample,
                                        row_sample=self.row_sample,
                                        seed=self.seed,
                                        backend=self.backend,
                                    ),
                                ),
                            ]
                        )

                    pipe.fit(X_train, y_train)
                    self.models[name] = pipe
                    y_pred = pipe.predict(X_test)
                    accuracy = accuracy_score(y_test, y_pred, normalize=True)
                    b_accuracy = balanced_accuracy_score(y_test, y_pred)
                    f1 = f1_score(y_test, y_pred, average="weighted")
                    try:
                        roc_auc = roc_auc_score(y_test, y_pred)
                    except Exception as exception:
                        roc_auc = None
                        if self.ignore_warnings is False:
                            print("ROC AUC couldn't be calculated for " + name)
                            print(exception)
                    names.append(name)
                    Accuracy.append(accuracy)
                    B_Accuracy.append(b_accuracy)
                    ROC_AUC.append(roc_auc)
                    F1.append(f1)
                    TIME.append(time.time() - start)
                    if self.custom_metric is not None:
                        custom_metric = self.custom_metric(y_test, y_pred)
                        CUSTOM_METRIC.append(custom_metric)
                    if self.verbose > 0:
                        if self.custom_metric is not None:
                            print(
                                {
                                    "Model": name,
                                    "Accuracy": accuracy,
                                    "Balanced Accuracy": b_accuracy,
                                    "ROC AUC": roc_auc,
                                    "F1 Score": f1,
                                    self.custom_metric.__name__: custom_metric,
                                    "Time taken": time.time() - start,
                                }
                            )
                        else:
                            print(
                                {
                                    "Model": name,
                                    "Accuracy": accuracy,
                                    "Balanced Accuracy": b_accuracy,
                                    "ROC AUC": roc_auc,
                                    "F1 Score": f1,
                                    "Time taken": time.time() - start,
                                }
                            )
                    if self.predictions:
                        predictions[name] = y_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        else:

            for name, model in tqdm(self.classifiers):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = CustomClassifier(
                            obj=model(random_state=self.random_state),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )
                    else:
                        pipe = CustomClassifier(
                            obj=model(),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            col_sample=self.col_sample,
                            row_sample=self.row_sample,
                            seed=self.seed,
                            backend=self.backend,
                        )

                    pipe.fit(X_train, y_train)
                    self.models[name] = pipe
                    y_pred = pipe.predict(X_test)
                    accuracy = accuracy_score(y_test, y_pred, normalize=True)
                    b_accuracy = balanced_accuracy_score(y_test, y_pred)
                    f1 = f1_score(y_test, y_pred, average="weighted")
                    try:
                        roc_auc = roc_auc_score(y_test, y_pred)
                    except Exception as exception:
                        roc_auc = None
                        if self.ignore_warnings is False:
                            print("ROC AUC couldn't be calculated for " + name)
                            print(exception)
                    names.append(name)
                    Accuracy.append(accuracy)
                    B_Accuracy.append(b_accuracy)
                    ROC_AUC.append(roc_auc)
                    F1.append(f1)
                    TIME.append(time.time() - start)
                    if self.custom_metric is not None:
                        custom_metric = self.custom_metric(y_test, y_pred)
                        CUSTOM_METRIC.append(custom_metric)
                    if self.verbose > 0:
                        if self.custom_metric is not None:
                            print(
                                {
                                    "Model": name,
                                    "Accuracy": accuracy,
                                    "Balanced Accuracy": b_accuracy,
                                    "ROC AUC": roc_auc,
                                    "F1 Score": f1,
                                    self.custom_metric.__name__: custom_metric,
                                    "Time taken": time.time() - start,
                                }
                            )
                        else:
                            print(
                                {
                                    "Model": name,
                                    "Accuracy": accuracy,
                                    "Balanced Accuracy": b_accuracy,
                                    "ROC AUC": roc_auc,
                                    "F1 Score": f1,
                                    "Time taken": time.time() - start,
                                }
                            )
                    if self.predictions:
                        predictions[name] = y_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        if self.custom_metric is None:
            scores = pd.DataFrame(
                {
                    "Model": names,
                    "Accuracy": Accuracy,
                    "Balanced Accuracy": B_Accuracy,
                    "ROC AUC": ROC_AUC,
                    "F1 Score": F1,
                    "Time Taken": TIME,
                }
            )
        else:
            scores = pd.DataFrame(
                {
                    "Model": names,
                    "Accuracy": Accuracy,
                    "Balanced Accuracy": B_Accuracy,
                    "ROC AUC": ROC_AUC,
                    "F1 Score": F1,
                    self.custom_metric.__name__: CUSTOM_METRIC,
                    "Time Taken": TIME,
                }
            )
        scores = scores.sort_values(
            by="Balanced Accuracy", ascending=False
        ).set_index("Model")

        if self.predictions:
            predictions_df = pd.DataFrame.from_dict(predictions)
        return scores, predictions_df if self.predictions is True else scores

    def provide_models(self, X_train, X_test, y_train, y_test):
        """
        This function returns all the model objects trained in fit function.
        If fit is not called already, then we call fit and then return the models.
        Parameters
        ----------
        X_train : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        X_test : array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.
        y_train : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y_test : array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.
        Returns
        -------
        models: dict-object,
            Returns a dictionary with each model pipeline as value 
            with key as name of models.
        """
        if len(self.models.keys()) == 0:
            self.fit(X_train, X_test, y_train, y_test)

        return self.models



"""Tools for MTS."""

# Authors: Thierry Moudiki <thierry.moudiki@gmail.com>
#
# License: BSD 3 Clear

import numpy as np
import pandas as pd

# compute output dates from data frame's index
def compute_input_dates(df):

    input_dates = df.index.values
        
    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))
        
    input_dates = pd.date_range(start=input_dates[0], 
                                periods=len(input_dates), 
                                freq=frequency).values.tolist()
        
    df_input_dates = pd.DataFrame({"date": input_dates})                    

    return pd.to_datetime(df_input_dates["date"]).dt.date


# compute output dates from data frame's index
def compute_output_dates(df, horizon):

    input_dates = df.index.values

    if input_dates[0] == 0:
        input_dates = pd.date_range(start = pd.Timestamp.today().strftime("%Y-%m-%d"), 
                                    periods=horizon)

    #print(f"\n in nnetsauce.utils.timeseries 1: {input_dates} \n")

    frequency = pd.infer_freq(pd.DatetimeIndex(input_dates))

    #print(f"\n in nnetsauce.utils.timeseries 2: {frequency} \n")

    output_dates = np.delete(
        pd.date_range(
            start=input_dates[-1], periods=horizon + 1, freq=frequency
        ).values,
        0,
    ).tolist()

    #print(f"\n in nnetsauce.utils.timeseries 3: {output_dates} \n")

    df_output_dates = pd.DataFrame({"date": output_dates})
    output_dates = pd.to_datetime(df_output_dates["date"]).dt.date

    return output_dates, frequency


# create lags for one series
def create_lags(x, k, n=None):

    k_ = k + 1

    n_k = len(x) - k_

    x_ = x[::-1]

    z = [x_[i : (n_k + i + 1)] for i in range(k_)]

    if n is None:

        return np.column_stack(z)

    temp = np.column_stack(z)
    return np.repeat(temp, n).reshape(temp.shape[0], -1)


# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# a = np.reshape(range(0, 24), (8, 3))
# create_train_inputs(a, 2)
def create_train_inputs(X, k):

    n_k = X.shape[0] - k

    z = [X[i : n_k + i, :] for i in range(1, (k + 1))]

    return (X[0:n_k, :], np.column_stack(z))


# reformat response in prediction loop
# a = np.reshape(range(0, 24), (8, 3))
# reformat_response(a, 2)
def reformat_response(X, k):

    return np.concatenate([X[i, :] for i in range(k)])


# create k lags for series x
# def create_lags(x, k):
#
#    n = len(x)
#    n_k = n - k
#    n_col = k + 1
#    res = np.zeros((n_k, n_col))
#
#    for i in range(n_k):
#        for j in range(n_col):
#            res[i, j] = x[i + j]
#
#    return res


# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# def create_train_inputs(X, k):
#
#    n, p = X.shape
#    n_k = n - k
#
#    y = np.zeros((n_k, p))
#    xreg = np.zeros((n_k, k*p))
#
#    for j in range(p):
#
#        Y_x = create_lags(X[:,j], k)
#
#        y[:,j] = Y_x[:, 0]
#
#        for l in range(1, k):
#
#            xreg[:, j*k + (l-1)] = Y_x[:, l]
#
#        xreg[:, j*k + (k-1)] = Y_x[:, k]
#
#    return (y, xreg)


# reformat response in prediction loop
# def reformat_response(X, k):
#
#    n, p = X.shape
#    res = np.zeros((1, k*p))
#
#    for j in range(p):
#        for i in range(k):
#            res[0, j*k + i] = X[i, j]
#
#    return res
#




# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
from ..utils.misc import flatten, is_factor

def dosubsample(y, row_sample=0.8, seed=123):
        
    index = []

    assert (row_sample < 1) & (
        row_sample >= 0
    ), "'row_sample' must be < 1 and >= 0"

    n_obs = len(y)
    n_obs_out = np.ceil(n_obs * row_sample)

    # preproc -----
    if is_factor(y): # classification

        classes, n_elem_classes = np.unique(y, return_counts=True)
        n_classes = len(classes)
        y_as_classes = y.copy()
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = float(n_elem_classes[i]) / n_obs

    else: # regression

        h = np.histogram(y, bins="auto")
        n_elem_classes = np.asarray(h[0], dtype=np.integer)
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = float(n_elem_classes[i]) / n_obs
        
        breaks = h[1]

        n_breaks_1 = len(breaks) - 1
        classes = range(n_breaks_1)
        n_classes = n_breaks_1
        y_as_classes = np.zeros_like(y, dtype=int)

        for i in classes:
            y_as_classes[(y > breaks[i]) * (y <= breaks[i + 1])] = int(i)

    # main loop ----

    np.random.seed(seed)

    for i in range(n_classes):

        bool_class_i = (y_as_classes == classes[i])               

        # index_class_i = [i for i, e in enumerate(bool_class_i) if e == True]
        index_class_i = np.asarray(np.where(bool_class_i == True)[0], dtype=np.integer)

        if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i

            index.append(
                np.random.choice(
                    index_class_i,
                    size=int(n_obs_out * freqs_hist[i]),  # output size
                    replace=True,
                ).tolist()
            )

        else:  # only one element in class

            try:

                index.append(index_class_i[0])

            except:

                0

    try:
        return np.asarray(flatten(index))
    except:
        return np.asarray(index)
    


import numpy as np

# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# check if x is int
def is_int(x):
    try:
        return int(x) == x
    except:
        return False


# check if x is float
def is_float(x):
    return isinstance(x, float)


# check if the response contains only integers
def is_factor(y):
    #return not np.mod(y, 1).any()
    return not np.array([is_float(y[i]) for i in range(len(y))]).all()


# flatten list of lists
#flatten = lambda l: [item for sublist in l for item in sublist]
def flatten(x): 
            
    res = []    
    
    for elt1 in x:
                    
        try:
            
            for elt2 in elt1:
                
                res.append(elt2)  
                
        except:
            
            res.append(elt1)  
                                
    return res




from .helpers import dosubsample

class SubSampler:
    """Subsampling class.

    Attributes:

       y: array-like, shape = [n_samples]
           Target values.

       row_sample: double
           subsampling fraction

       seed: int
           reproductibility seed

    """

    def __init__(self, y, row_sample=0.8, seed=123):
        self.y = y
        self.row_sample = row_sample
        self.seed = seed

    def subsample(self):
        return dosubsample(self.y, self.row_sample, self.seed)
